import multiprocessing as mp
import numpy as np
import time

# How many CPU's are available:
print("Number of processors: ", mp.cpu_count())
USED_CPUS = 10
print(f"Using {USED_CPUS} CPUs")


# ----- SAMPLE PROBLEM: Return how many numbers per row are inside the requested Range -----

# Prepare Example data
np.random.RandomState(100)
arr = np.random.randint(0, 10, size=[2000000, 5]) # Matrix with 2 million rows
data = arr.tolist()

# Main Function
def howmany_within_range(i, row, minimum, maximum):
    """Returns how many numbers lie within `maximum` and `minimum` in a given `row`"""
    count = 0
    for n in row:
        if minimum <= n <= maximum:
            count = count + 1
    return i, count

# Define callback function to collect the output in `results`
def collect_result(result):
    global results
    results.append(result)

if __name__ == '__main__':
    # Solution Without Paralleization
    start = time.time()
    results = []
    for i, row in enumerate(data):
        results.append(howmany_within_range(i, row, minimum=4, maximum=8))
    end = time.time()
    print(results[:10])
    print(f"SEQUENTIAL: It took {(end - start)} seconds!")
    
    # Parallelizing Solution
    start = time.time()
    results = []
    with mp.Pool(processes=USED_CPUS) as pool:
        for i, row in enumerate(data):
            res = pool.apply_async(howmany_within_range, args=(i, row, 4, 8), callback=collect_result)
        try:
            print(res.get(timeout=1))
        except mp.TimeoutError:
            print("We lacked patience and got a multiprocessing.TimeoutError")
        print("For the moment, the pool remains available for more work")
    
    end = time.time()
    print(results[:10])
    print(f"PARALLEL: It took {(end - start)} seconds!")
